import copy
import logging
from dataclasses import asdict, dataclass, field


def default_field(obj):
    return field(default_factory=lambda: copy.copy(obj))


class ConfigInterface(object):

    ### global logging-config setup
    # Only do basicConfig if no handlers have been configured
    LOG_PATH: str = "./"
    LOG_LEVEL: ... = logging.INFO
    LOG_FORMAT: str = "%(asctime)-15s: %(name)-32s - %(message)s"

    def __post_init__(self):
        if len(logging._handlerList) == 0:
            logging.basicConfig(level=self.LOG_LEVEL, format=self.LOG_FORMAT)

    def toDict(self):
        result = asdict(self)
        delkeys = [k for k in result.keys() if k.startswith("LOG_")]
        for key in delkeys:
            del result[key]
        return result

    @classmethod
    def fromDict(cls, config_dict):
        return cls(**config_dict)



from pymongo import MongoClient


class QueueFactory:

    QUEUE_METHOD_FAKE = "fake"
    QUEUE_METHOD_MONGODB = "mongodb"

    @staticmethod
    def getQueue(config, storage=None, consumer_id=None):
        queue_config = config.QUEUE_CONFIG
        if queue_config.QUEUE_METHOD == QueueFactory.QUEUE_METHOD_FAKE:
            from mcrit.queue.LocalQueue import LocalQueue
            from mcrit.Worker import Worker

            if storage is None:
                raise ValueError("QueueFactory needs storage to set up LocalQueue")
            queue = LocalQueue()
            queue.set_worker(Worker(queue=queue, storage=storage, config=config))
        else:
            from mcrit.libs.mongoqueue import MongoQueue

            if consumer_id is None:
                raise ValueError("QueueFactory needs consumer_id to set up MongoQueue")
            db = MongoClient(host=queue_config.QUEUE_SERVER, port=queue_config.QUEUE_PORT)
            collection = db[queue_config.QUEUE_MONGODB_DBNAME].queue
            queue = MongoQueue(
                collection,
                consumer_id,
                timeout=queue_config.QUEUE_TIMEOUT,
                max_attempts=queue_config.QUEUE_MAX_ATTEMPTS,
            )
        queue.clean_interval = queue_config.QUEUE_CLEAN_INTERVAL
        return queue



import functools
import time
import json
import logging
from typing import Dict, List, Optional, Tuple

import requests
import urllib.parse
from mcrit.storage.FamilyEntry import FamilyEntry
from mcrit.storage.FunctionEntry import FunctionEntry
from mcrit.storage.SampleEntry import SampleEntry
from mcrit.queue.LocalQueue import Job
from smda.common.SmdaReport import SmdaReport
from smda.Disassembler import Disassembler


# Only do basicConfig if no handlers have been configured
if len(logging._handlerList) == 0:
    logging.basicConfig(level=logging.INFO, format="%(asctime)-15s %(message)s")
LOGGER = logging.getLogger(__name__)


class JobTerminatedError(Exception):
    pass


def isJobTerminated(job):
    if job is None:
        return True

    return job.is_terminated

def isJobFailed(job):
    return (job is not None) and (not job.is_failed)

def isJobFinishedTerminatedOrFailed(job):
    return isJobTerminated(job) or (job.result is not None) or isJobFailed(job)


def handle_response(response):
    data = None
    if response.status_code in [500, 501]:
        LOGGER.warning("McritClient received status code 500 from MCRIT.")
    elif response.status_code in [400, 404, 410]:
        # nothing to here as of now
        pass
    elif response.status_code in [200, 202]:
        json_response = response.json()
        if "status" in json_response and json_response["status"] == "successful":
            data = json_response["data"]
    return data


class McritClient:
    def __init__(self, mcrit_server=None):
        self.mcrit_server = "http://localhost:8000"
        if mcrit_server is not None:
            self.mcrit_server = mcrit_server

    def _getMatchingRequestParams(
        self, minhash_threshold, pichash_size, force_recalculation, band_matches_required
    ):
        params = {}
        if minhash_threshold is not None:
            params["minhash_score"] = minhash_threshold
        if pichash_size is not None:
            params["pichash_size"] = pichash_size
        if force_recalculation is not None:
            params["force_recalculation"] = force_recalculation
        if band_matches_required is not None:
            params["band_matches_required"] = band_matches_required
        return params

    def respawn(self):
        response = requests.post(f"{self.mcrit_server}/respawn")
        return handle_response(response)

    def addReport(self, smda_report: SmdaReport) -> Tuple[SampleEntry, Optional[str]]:
        smda_json = smda_report.toDict()
        response = requests.post(f"{self.mcrit_server}/samples", json=smda_json)
        data = handle_response(response)
        if data is not None:
            if "job_id" in data:
                job_id = data["job_id"]
            else:
                job_id = None
            return SampleEntry.fromDict(data["sample_info"]), job_id

    def addBinarySample(self, binary: bytes, filename=None, family=None, version=None, is_dump=False, base_addr=None, bitness=None) -> Tuple[SampleEntry, Optional[str]]:
        query_fields = []
        if filename is not None:
            query_fields.append(f"filename={filename}")
        if family is not None:
            query_fields.append(f"family={family}")
        if version is not None:
            query_fields.append(f"version={version}")
        if is_dump:
            query_fields.append(f"is_dump=1")
        if base_addr is not None:
            query_fields.append(f"base_addr=0x{base_addr:x}")
        if bitness is not None and bitness in [32, 64]:
            query_fields.append(f"bitness={bitness}")
        query_string = ""
        if len(query_fields) > 0:
            query_string = "?" + "&".join(query_fields)
        response = requests.post(f"{self.mcrit_server}/samples/binary{query_string}", data=binary)
        return handle_response(response)

    ###########################################
    ### Families 
    ###########################################

    def modifyFamily(self, family_id, family_name=None, is_library=None):
        update_dict = {}
        if family_name is not None:
            update_dict["family_name"] = family_name
        if is_library is not None:
            update_dict["is_library"] = is_library
        response = requests.put(f"{self.mcrit_server}/families/{family_id}", update_dict)
        return handle_response(response)

    def getFamily(self, family_id: int, with_samples=True) -> Optional[FamilyEntry]:
        query_params = "?with_samples=true" if with_samples else "?with_samples=false"
        response = requests.get(f"{self.mcrit_server}/families/{family_id}{query_params}")
        data = handle_response(response)
        if data is not None:
            return FamilyEntry.fromDict(data)
        return None

    def getFamilies(self) -> Optional[Dict[int, FamilyEntry]]:
        response = requests.get(f"{self.mcrit_server}/families")
        data = handle_response(response)
        if data is not None:
            return {i: FamilyEntry.fromDict(entry) for i, entry in data.items()}
        return None

    def isFamilyId(self, family_id) -> bool:
        return self.getFamily(family_id, with_samples=False) is not None

    def deleteFamily(self, family_id, keep_samples=False):
        query_params = "?keep_samples=true" if keep_samples else "?keep_samples=false"
        response = requests.delete(f"{self.mcrit_server}/families/{family_id}{query_params}")
        return handle_response(response)

    ###########################################
    ### Samples 
    ###########################################

    def isSampleId(self, sample_id):
        response = requests.get(f"{self.mcrit_server}/samples/{sample_id}")
        data = handle_response(response)
        if data is not None:
            return True
        return False

    def modifySample(self, sample_id, family_name=None, version=None, component=None, is_library=None):
        update_dict = {}
        if family_name is not None:
            update_dict["family_name"] = family_name
        if version is not None:
            update_dict["version"] = version
        if component is not None:
            update_dict["component"] = component
        if is_library is not None:
            update_dict["is_library"] = is_library
        response = requests.put(f"{self.mcrit_server}/samples/{sample_id}", update_dict)
        return handle_response(response)

    def deleteSample(self, sample_id):
        response = requests.delete(f"{self.mcrit_server}/samples/{sample_id}")
        return handle_response(response)

    def getSamplesByFamilyId(self, family_id: int) -> Optional[List[SampleEntry]]:
        family_data = self.getFamily(family_id)
        if family_data is not None:
            return family_data.samples

    def getSampleById(self, sample_id):
        response = requests.get(f"{self.mcrit_server}/samples/{sample_id}")
        data = handle_response(response)
        if data is not None:
            return SampleEntry.fromDict(data)

    def getSamples(self, start=0, limit=0):
        query_string = ""
        if (isinstance(start, int) and start >= 0) and (isinstance(limit, int) and limit >= 0):
            query_string = f"?start={start}&limit={limit}"
        response = requests.get(f"{self.mcrit_server}/samples{query_string}")
        data = handle_response(response)
        if data is not None:
            return {int(k): SampleEntry.fromDict(v) for k, v in data.items()}

    ###########################################
    ### Functions
    ###########################################

    def getFunctionsBySampleId(self, sample_id):
        response = requests.get(f"{self.mcrit_server}/samples/{sample_id}/functions")
        data = handle_response(response)
        if data is not None:
            return [
                FunctionEntry.fromDict(function_entry_dict)
                for function_entry_dict in data.values()
            ]

    def getFunctions(self, start=0, limit=0):
        query_string = ""
        if (isinstance(start, int) and start >= 0) and (isinstance(limit, int) and limit >= 0):
            query_string = f"?start={start}&limit={limit}"
        response = requests.get(f"{self.mcrit_server}/functions{query_string}")
        data = handle_response(response)
        if data is not None:
            return {int(k): FunctionEntry.fromDict(v) for k, v in data.items()}

    def isFunctionId(self, function_id):
        response = requests.get(f"{self.mcrit_server}/functions/{function_id}")
        data = handle_response(response)
        if data is not None:
            return True
        return False

    def getFunctionById(self, function_id: int, with_xcfg=False) -> Optional[FunctionEntry]:
        query_with_xcfg = "?with_xcfg=True" if with_xcfg else ""
        response = requests.get(f"{self.mcrit_server}/functions/{function_id}{query_with_xcfg}")
        data = handle_response(response)
        if data is not None:
            return FunctionEntry.fromDict(data)

    def completeMinhashes(self):
        response = requests.get(f"{self.mcrit_server}/complete_minhashes")
        return handle_response(response)

    ###########################################
    ### Matching 
    ###########################################

    def requestMatchesForSmdaReport(
        self,
        smda_report: SmdaReport,
        minhash_threshold=None,
        pichash_size=None,
        band_matches_required=None,
        force_recalculation=False,
    ) -> str:
        smda_json = smda_report.toDict()
        params = self._getMatchingRequestParams(
            minhash_threshold, pichash_size, force_recalculation, band_matches_required
        )
        response = requests.post(f"{self.mcrit_server}/query", json=smda_json, params=params)
        return handle_response(response)

    def requestMatchesForMappedBinary(
        self,
        binary: bytes,
        base_address: int,
        minhash_threshold=None,
        pichash_size=None,
        band_matches_required=None,
        disassemble_locally=True,
        force_recalculation=False,
    ) -> str:
        if disassemble_locally:
            disassembler = Disassembler()
            smda_report = disassembler.disassembleBuffer(binary, base_address)
            if smda_report.status == "error":
                return None
            return self.requestMatchesForSmdaReport(
                smda_report,
                minhash_threshold=minhash_threshold,
                pichash_size=pichash_size,
                band_matches_required=band_matches_required,
                force_recalculation=force_recalculation,
            )

        params = self._getMatchingRequestParams(
            minhash_threshold, pichash_size, force_recalculation, band_matches_required
        )
        response = requests.post(f"{self.mcrit_server}/query/binary/mapped/{base_address}", binary, params=params)
        return handle_response(response)

    def requestMatchesForUnmappedBinary(
        self,
        binary: bytes,
        minhash_threshold=None,
        pichash_size=None,
        band_matches_required=None,
        disassemble_locally=True,
        force_recalculation=False,
    ) -> str:
        if disassemble_locally:
            disassembler = Disassembler()
            smda_report = disassembler.disassembleUnmappedBuffer(binary)
            if smda_report.status == "error":
                return None
            return self.requestMatchesForSmdaReport(
                smda_report,
                minhash_threshold=minhash_threshold,
                pichash_size=pichash_size,
                band_matches_required=band_matches_required,
                force_recalculation=force_recalculation,
            )

        params = self._getMatchingRequestParams(
            minhash_threshold, pichash_size, force_recalculation, band_matches_required
        )

        response = requests.post(f"{self.mcrit_server}/query/binary", binary, params=params)
        return handle_response(response)

    def requestMatchesForSample(
        self,
        sample_id,
        minhash_threshold=None,
        pichash_size=None,
        band_matches_required=None,
        force_recalculation=False,
    ) -> None:
        params = self._getMatchingRequestParams(
            minhash_threshold, pichash_size, force_recalculation, band_matches_required
        )
        response = requests.get(f"{self.mcrit_server}/matches/sample/{sample_id}", params=params)
        return handle_response(response)

    def requestMatchesForSampleVs(
        self,
        sample_id,
        other_sample_id,
        minhash_threshold=None,
        pichash_size=None,
        band_matches_required=None,
        force_recalculation=False,
    ) -> str:
        params = self._getMatchingRequestParams(
            minhash_threshold, pichash_size, force_recalculation, band_matches_required
        )
        response = requests.get(f"{self.mcrit_server}/matches/sample/{sample_id}/{other_sample_id}",params=params)
        return handle_response(response)

    def requestMatchesCross(
        self,
        sample_ids,
        minhash_threshold=None,
        pichash_size=None,
        band_matches_required=None,
        force_recalculation=False,
    ) -> None:
        params = self._getMatchingRequestParams(
            minhash_threshold, pichash_size, force_recalculation, band_matches_required
        )
        response = requests.get(
            f"{self.mcrit_server}/matches/sample/cross/{','.join([str(id) for id in sample_ids])}",
            params=params
        )
        return handle_response(response)

    def getMatchFunctionVs(
        self,
        function_id_a:int,
        function_id_b:int
    ) -> None:
        response = requests.get(f"{self.mcrit_server}/matches/function/{function_id_a}/{function_id_b}")
        return handle_response(response)

    def getMatchesForPicHash(self, pichash, summary=False):
        summary_string = "/summary" if summary else ""
        response = requests.get(f"{self.mcrit_server}/query/pichash/{pichash:016x}{summary_string}")
        return handle_response(response)

    def getMatchesForPicBlockHash(self, picblockhash, summary=False):
        summary_string = "/summary" if summary else ""
        response = requests.get(f"{self.mcrit_server}/query/picblockhash/{picblockhash:016x}{summary_string}")
        return handle_response(response)

    def getSampleBySha256(self, sample_sha256: str):
        response = requests.get(f"{self.mcrit_server}/samples/sha256/{sample_sha256}")
        data = handle_response(response)
        if data is None:
            return None
        return SampleEntry.fromDict(data)

    ###########################################
    ### Status, Results 
    ###########################################

    def getStatus(self):
        response = requests.get(f"{self.mcrit_server}/status")
        return handle_response(response)

    def getVersion(self):
        response = requests.get(f"{self.mcrit_server}/version")
        data = handle_response(response)
        if isinstance(data, dict) and "version" in data:
            return data["version"]
        return None

    def getJobCount(self, filter=None):
        query_string = ""
        if isinstance(filter, str) and filter is not None:
            if len(query_string) == 0:
                query_string = f"?filter={filter}"
            else:
                query_string += f"&filter={filter}"
        response = requests.get(f"{self.mcrit_server}/jobs/{query_string}")
        data = handle_response(response)
        if data is not None:
            return len(data)

    def getQueueData(self, start=0, limit=0, filter=None):
        query_string = ""
        if isinstance(start, int) and start > 0:
            if len(query_string) == 0:
                query_string = f"?start={start}"
            else:
                query_string += f"&start={start}"
        if isinstance(limit, int) and limit > 0:
            if len(query_string) == 0:
                query_string = f"?limit={limit}"
            else:
                query_string += f"&limit={limit}"
        if isinstance(filter, str) and filter is not None:
            if len(query_string) == 0:
                query_string = f"?filter={filter}"
            else:
                query_string += f"&filter={filter}"
        response = requests.get(f"{self.mcrit_server}/jobs/{query_string}")
        data = handle_response(response)
        if data is not None:
            return [Job(job_data, None) for job_data in data]

    def getJobData(self, job_id):
        response = requests.get(f"{self.mcrit_server}/jobs/{job_id}")
        data = handle_response(response)
        if data is not None:
            return Job(data, None)

    def getResultForJob(self, job_id):
        response = requests.get(f"{self.mcrit_server}/jobs/{job_id}/result")
        return handle_response(response)

    def getResult(self, result_id):
        response = requests.get(f"{self.mcrit_server}/results/{result_id}")
        return handle_response(response)

    def getJobForResult(self, result_id):
        response = requests.get(f"{self.mcrit_server}/results/{result_id}/job")
        data = handle_response(response)
        if data is not None:
            return Job(data, None)

    def awaitResult(self, job_id, sleep_time=2):
        if job_id is None:
            return None
        job = self.getJobData(job_id)
        while not isJobFinishedTerminatedOrFailed(job):
            time.sleep(sleep_time)
            job = self.getJobData(job_id)
        if isJobTerminated(job):
            raise JobTerminatedError
        result_id = job.result
        return self.getResult(result_id)

    ###########################################
    ### Import / Export
    ###########################################

    def getExportData(self, sample_ids=None, compress_data=True) -> dict:
        compress_uri_param = "?compress=True" if compress_data else ""
        result_data = {}
        if sample_ids is not None:
            if isinstance(sample_ids, list) and all(isinstance(item, int) for item in sample_ids):
                sample_ids_as_str = ",".join([str(sample_id) for sample_id in sample_ids])
                response = requests.get(f"{self.mcrit_server}/export/{sample_ids_as_str}{compress_uri_param}")
                result_data = handle_response(response)
            else:
                raise ValueError("sample_ids must be a list of int.")
        else:
            response = requests.get(f"{self.mcrit_server}/export{compress_uri_param}")
            result_data = handle_response(response)
        return result_data

    def addImportData(self, import_data):
        if not isinstance(import_data, dict):
            raise ValueError("Can only forward dictionaries with export data.")
        response = requests.post(f"{self.mcrit_server}/import", json=import_data)
        return handle_response(response)

    ###########################################
    ### Unique Blocks
    ###########################################

    def requestUniqueBlocksForSamples(self, sample_ids: List[int]) -> Dict:
        if isinstance(sample_ids, list) and all(isinstance(item, int) for item in sample_ids):
            sample_ids_as_str = ",".join([str(sample_id) for sample_id in sample_ids])
            response = requests.get(f"{self.mcrit_server}/uniqueblocks/samples/{sample_ids_as_str}")
            result_data = handle_response(response)
        else:
            raise ValueError("sample_ids must be a list of int.")
        return result_data

    def requestUniqueBlocksForFamily(self, family_id: int) -> Dict:
        if isinstance(family_id, int):
            response = requests.get(f"{self.mcrit_server}/uniqueblocks/family/{family_id}")
            result_data = handle_response(response)
        else:
            raise ValueError("sample_ids must be a list of int.")
        return result_data

    ###########################################
    ### Search
    ###########################################

    # When performing an initial search, the cursor should be set to None.
    # Search results are of the following form:
    # {
    #     "search_results": {
    #         id1: found_entry1,
    #         id2: found_entry2,
    #         ...
    #     },
    #     "cursor": {
    #         "forward": forward cursor,
    #         "backward": backward cursor,
    #     }
    # }
    # To get further results, perform a search using the forward cursor.
    # To get back to the previous search results, use the backward cursor.
    # If no further or previous results are available, the forward or backward cursor will be None.
    #
    # IMPORTANT: A cursor shall only be used in combination with the same
    # search_term, is_ascending and sort_by value that were used when the cursor was returned from mcrit.
    # If those parameters are altered, mcrit's behavior is undefined.
    
    def _search_base(self, search_kind, search_term, cursor=None, is_ascending=True, sort_by=None, limit=None):
        params = {
            "query": search_term,
            "is_ascending": is_ascending,
        }
        if cursor is not None:
            params["cursor"] = cursor
        if sort_by is not None:
            params["sort_by"] = sort_by 
        if limit is not None:
            params["limit"] = limit 
        encoded_params = urllib.parse.urlencode(params)
        response = requests.get(f"{self.mcrit_server}/search/{search_kind}?{encoded_params}")
        return handle_response(response)

    search_families = functools.partialmethod(_search_base, "families")

    search_samples = functools.partialmethod(_search_base, "samples")

    search_functions = functools.partialmethod(_search_base, "functions")



import hashlib
import logging
from dataclasses import asdict, dataclass, field

from mcrit.config.ConfigInterface import ConfigInterface, default_field
from mcrit.minhash.MinHasher import MinHasher


@dataclass
class MinHashConfig(ConfigInterface):
    # Store the combination of shingles (unsorted) that were used to create the given MinHash
    MINHASH_TRACK_SHINGLES: bool = False
    # Which minhash calculation strategy should be used
    MINHASH_STRATEGY: ... = MinHasher.MINHASH_STRATEGY_SEGMENTED
    # A function must consist of minimum N instructions to be considered for MinHashing
    MINHASH_FN_MIN_INS: int = 10
    # A function must alternatively consist of minimum N basic blocks to be considered for MinHashing
    MINHASH_FN_MIN_BLOCKS: int = 0
    # Length in number of Shingles of which a minhash consists
    MINHASH_SIGNATURE_LENGTH: int = 64
    # Number of bits per signature element (1-32 bits)
    MINHASH_SIGNATURE_BITS: int = 8
    # The lower bound at which paired MinHashes are considered a match (range: 0-100)
    MINHASH_MATCHING_THRESHOLD: int = 50
    # random seed to be used when initiating XOR values for minhash seeds
    MINHASH_SEED: int = 0xDEADBEEF
    # When using as server, Gunicorn/Falcon may have issues with multiprocessing while indexing, which can be disabled this way.
    MINHASH_POOL_INDEXING: bool = True
    MINHASH_POOL_MATCHING: bool = True
    # The minimum number of band matches a minhash must have before being considered a candidate for matching
    BAND_MATCHES_REQUIRED = 1
    # minimum function size for considering PicHash matching
    PICHASH_SIZE: int = 10
    # do not perform minhash matching for pichash matches, instead assume they are implied
    PICHASH_IMPLIES_MINHASH_MATCH: bool = True

    def getConfigHash(self):
        config_str = ""
        config_str += f"_{self.MINHASH_STRATEGY}_{self.MINHASH_FN_MIN_INS}_{self.MINHASH_FN_MIN_BLOCKS}"
        config_str += f"_{self.MINHASH_SIGNATURE_LENGTH}_{self.MINHASH_SIGNATURE_BITS}_{self.MINHASH_SEED}"
        config_str += f"_{self.PICHASH_SIZE}"
        return hashlib.sha256(config_str.encode('utf-8')).hexdigest()



'''
pymmh3 was written by Fredrik Kihlander and enhanced by Swapnil Gusani, and is placed in the public
domain. The authors hereby disclaim copyright to this source code.

pure python implementation of the murmur3 hash algorithm

https://code.google.com/p/smhasher/wiki/MurmurHash3

This was written for the times when you do not want to compile c-code and install modules,
and you only want a drop-in murmur3 implementation.

As this is purely python it is FAR from performant and if performance is anything that is needed
a proper c-module is suggested!

This module is written to have the same format as mmh3 python package found here for simple conversions:

https://pypi.python.org/pypi/mmh3/2.3.1
'''

import sys as _sys
if (_sys.version_info > (3, 0)):
    def xrange( a, b, c ):
        return range( a, b, c )
    def xencode(x):
        if isinstance(x, bytes) or isinstance(x, bytearray):
            return x
        else:
            return x.encode()
else:
    def xencode(x):
        return x
del _sys

def hash( key, seed = 0x0 ):
    ''' Implements 32bit murmur3 hash. '''

    key = bytearray( xencode(key) )

    def fmix( h ):
        h ^= h >> 16
        h  = ( h * 0x85ebca6b ) & 0xFFFFFFFF
        h ^= h >> 13
        h  = ( h * 0xc2b2ae35 ) & 0xFFFFFFFF
        h ^= h >> 16
        return h

    length = len( key )
    nblocks = int( length / 4 )

    h1 = seed

    c1 = 0xcc9e2d51
    c2 = 0x1b873593

    # body
    for block_start in xrange( 0, nblocks * 4, 4 ):
        # ??? big endian?
        k1 = key[ block_start + 3 ] << 24 | \
             key[ block_start + 2 ] << 16 | \
             key[ block_start + 1 ] <<  8 | \
             key[ block_start + 0 ]
             
        k1 = ( c1 * k1 ) & 0xFFFFFFFF
        k1 = ( k1 << 15 | k1 >> 17 ) & 0xFFFFFFFF # inlined ROTL32
        k1 = ( c2 * k1 ) & 0xFFFFFFFF
        
        h1 ^= k1
        h1  = ( h1 << 13 | h1 >> 19 ) & 0xFFFFFFFF # inlined ROTL32
        h1  = ( h1 * 5 + 0xe6546b64 ) & 0xFFFFFFFF

    # tail
    tail_index = nblocks * 4
    k1 = 0
    tail_size = length & 3

    if tail_size >= 3:
        k1 ^= key[ tail_index + 2 ] << 16
    if tail_size >= 2:
        k1 ^= key[ tail_index + 1 ] << 8
    if tail_size >= 1:
        k1 ^= key[ tail_index + 0 ]
    
    if tail_size > 0:
        k1  = ( k1 * c1 ) & 0xFFFFFFFF
        k1  = ( k1 << 15 | k1 >> 17 ) & 0xFFFFFFFF # inlined ROTL32
        k1  = ( k1 * c2 ) & 0xFFFFFFFF
        h1 ^= k1

    #finalization
    unsigned_val = fmix( h1 ^ length )
    if unsigned_val & 0x80000000 == 0:
        return unsigned_val
    else:
        return -( (unsigned_val ^ 0xFFFFFFFF) + 1 )


def hash128( key, seed = 0x0, x64arch = True ):
    ''' Implements 128bit murmur3 hash. '''
    def hash128_x64( key, seed ):
        ''' Implements 128bit murmur3 hash for x64. '''

        def fmix( k ):
            k ^= k >> 33
            k  = ( k * 0xff51afd7ed558ccd ) & 0xFFFFFFFFFFFFFFFF
            k ^= k >> 33
            k  = ( k * 0xc4ceb9fe1a85ec53 ) & 0xFFFFFFFFFFFFFFFF
            k ^= k >> 33
            return k

        length = len( key )
        nblocks = int( length / 16 )

        h1 = seed
        h2 = seed

        c1 = 0x87c37b91114253d5
        c2 = 0x4cf5ad432745937f

        #body
        for block_start in xrange( 0, nblocks * 8, 8 ):
            # ??? big endian?
            k1 = key[ 2 * block_start + 7 ] << 56 | \
                 key[ 2 * block_start + 6 ] << 48 | \
                 key[ 2 * block_start + 5 ] << 40 | \
                 key[ 2 * block_start + 4 ] << 32 | \
                 key[ 2 * block_start + 3 ] << 24 | \
                 key[ 2 * block_start + 2 ] << 16 | \
                 key[ 2 * block_start + 1 ] <<  8 | \
                 key[ 2 * block_start + 0 ]

            k2 = key[ 2 * block_start + 15 ] << 56 | \
                 key[ 2 * block_start + 14 ] << 48 | \
                 key[ 2 * block_start + 13 ] << 40 | \
                 key[ 2 * block_start + 12 ] << 32 | \
                 key[ 2 * block_start + 11 ] << 24 | \
                 key[ 2 * block_start + 10 ] << 16 | \
                 key[ 2 * block_start + 9 ] <<  8 | \
                 key[ 2 * block_start + 8 ]

            k1  = ( c1 * k1 ) & 0xFFFFFFFFFFFFFFFF
            k1  = ( k1 << 31 | k1 >> 33 ) & 0xFFFFFFFFFFFFFFFF # inlined ROTL64
            k1  = ( c2 * k1 ) & 0xFFFFFFFFFFFFFFFF
            h1 ^= k1

            h1 = ( h1 << 27 | h1 >> 37 ) & 0xFFFFFFFFFFFFFFFF # inlined ROTL64
            h1 = ( h1 + h2 ) & 0xFFFFFFFFFFFFFFFF
            h1 = ( h1 * 5 + 0x52dce729 ) & 0xFFFFFFFFFFFFFFFF

            k2  = ( c2 * k2 ) & 0xFFFFFFFFFFFFFFFF
            k2  = ( k2 << 33 | k2 >> 31 ) & 0xFFFFFFFFFFFFFFFF # inlined ROTL64
            k2  = ( c1 * k2 ) & 0xFFFFFFFFFFFFFFFF
            h2 ^= k2

            h2 = ( h2 << 31 | h2 >> 33 ) & 0xFFFFFFFFFFFFFFFF # inlined ROTL64
            h2 = ( h1 + h2 ) & 0xFFFFFFFFFFFFFFFF
            h2 = ( h2 * 5 + 0x38495ab5 ) & 0xFFFFFFFFFFFFFFFF

        #tail
        tail_index = nblocks * 16
        k1 = 0
        k2 = 0
        tail_size = length & 15

        if tail_size >= 15:
            k2 ^= key[ tail_index + 14 ] << 48
        if tail_size >= 14:
            k2 ^= key[ tail_index + 13 ] << 40
        if tail_size >= 13:
            k2 ^= key[ tail_index + 12 ] << 32
        if tail_size >= 12:
            k2 ^= key[ tail_index + 11 ] << 24
        if tail_size >= 11:
            k2 ^= key[ tail_index + 10 ] << 16
        if tail_size >= 10:
            k2 ^= key[ tail_index +  9 ] << 8
        if tail_size >=  9:
            k2 ^= key[ tail_index +  8 ]

        if tail_size > 8:
            k2  = ( k2 * c2 ) & 0xFFFFFFFFFFFFFFFF
            k2  = ( k2 << 33 | k2 >> 31 ) & 0xFFFFFFFFFFFFFFFF # inlined ROTL64
            k2  = ( k2 * c1 ) & 0xFFFFFFFFFFFFFFFF
            h2 ^= k2

        if tail_size >=  8:
            k1 ^= key[ tail_index +  7 ] << 56
        if tail_size >=  7:
            k1 ^= key[ tail_index +  6 ] << 48
        if tail_size >=  6:
            k1 ^= key[ tail_index +  5 ] << 40
        if tail_size >=  5:
            k1 ^= key[ tail_index +  4 ] << 32
        if tail_size >=  4:
            k1 ^= key[ tail_index +  3 ] << 24
        if tail_size >=  3:
            k1 ^= key[ tail_index +  2 ] << 16
        if tail_size >=  2:
            k1 ^= key[ tail_index +  1 ] << 8
        if tail_size >=  1:
            k1 ^= key[ tail_index +  0 ]

        if tail_size > 0:
            k1  = ( k1 * c1 ) & 0xFFFFFFFFFFFFFFFF
            k1  = ( k1 << 31 | k1 >> 33 ) & 0xFFFFFFFFFFFFFFFF # inlined ROTL64
            k1  = ( k1 * c2 ) & 0xFFFFFFFFFFFFFFFF
            h1 ^= k1

        #finalization
        h1 ^= length
        h2 ^= length

        h1  = ( h1 + h2 ) & 0xFFFFFFFFFFFFFFFF
        h2  = ( h1 + h2 ) & 0xFFFFFFFFFFFFFFFF

        h1  = fmix( h1 )
        h2  = fmix( h2 )

        h1  = ( h1 + h2 ) & 0xFFFFFFFFFFFFFFFF
        h2  = ( h1 + h2 ) & 0xFFFFFFFFFFFFFFFF

        return ( h2 << 64 | h1 )

    def hash128_x86( key, seed ):
        ''' Implements 128bit murmur3 hash for x86. '''

        def fmix( h ):
            h ^= h >> 16
            h  = ( h * 0x85ebca6b ) & 0xFFFFFFFF
            h ^= h >> 13
            h  = ( h * 0xc2b2ae35 ) & 0xFFFFFFFF
            h ^= h >> 16
            return h

        length = len( key )
        nblocks = int( length / 16 )

        h1 = seed
        h2 = seed
        h3 = seed
        h4 = seed

        c1 = 0x239b961b
        c2 = 0xab0e9789
        c3 = 0x38b34ae5
        c4 = 0xa1e38b93

        #body
        for block_start in xrange( 0, nblocks * 16, 16 ):
            k1 = key[ block_start +  3 ] << 24 | \
                 key[ block_start +  2 ] << 16 | \
                 key[ block_start +  1 ] <<  8 | \
                 key[ block_start +  0 ]

            k2 = key[ block_start +  7 ] << 24 | \
                 key[ block_start +  6 ] << 16 | \
                 key[ block_start +  5 ] <<  8 | \
                 key[ block_start +  4 ]

            k3 = key[ block_start + 11 ] << 24 | \
                 key[ block_start + 10 ] << 16 | \
                 key[ block_start +  9 ] <<  8 | \
                 key[ block_start +  8 ]

            k4 = key[ block_start + 15 ] << 24 | \
                 key[ block_start + 14 ] << 16 | \
                 key[ block_start + 13 ] <<  8 | \
                 key[ block_start + 12 ]

            k1  = ( c1 * k1 ) & 0xFFFFFFFF
            k1  = ( k1 << 15 | k1 >> 17 ) & 0xFFFFFFFF # inlined ROTL32
            k1  = ( c2 * k1 ) & 0xFFFFFFFF
            h1 ^= k1

            h1 = ( h1 << 19 | h1 >> 13 ) & 0xFFFFFFFF # inlined ROTL32
            h1 = ( h1 + h2 ) & 0xFFFFFFFF
            h1 = ( h1 * 5 + 0x561ccd1b ) & 0xFFFFFFFF

            k2  = ( c2 * k2 ) & 0xFFFFFFFF
            k2  = ( k2 << 16 | k2 >> 16 ) & 0xFFFFFFFF # inlined ROTL32
            k2  = ( c3 * k2 ) & 0xFFFFFFFF
            h2 ^= k2

            h2 = ( h2 << 17 | h2 >> 15 ) & 0xFFFFFFFF # inlined ROTL32
            h2 = ( h2 + h3 ) & 0xFFFFFFFF
            h2 = ( h2 * 5 + 0x0bcaa747 ) & 0xFFFFFFFF

            k3  = ( c3 * k3 ) & 0xFFFFFFFF
            k3  = ( k3 << 17 | k3 >> 15 ) & 0xFFFFFFFF # inlined ROTL32
            k3  = ( c4 * k3 ) & 0xFFFFFFFF
            h3 ^= k3

            h3 = ( h3 << 15 | h3 >> 17 ) & 0xFFFFFFFF # inlined ROTL32
            h3 = ( h3 + h4 ) & 0xFFFFFFFF
            h3 = ( h3 * 5 + 0x96cd1c35 ) & 0xFFFFFFFF

            k4  = ( c4 * k4 ) & 0xFFFFFFFF
            k4  = ( k4 << 18 | k4 >> 14 ) & 0xFFFFFFFF # inlined ROTL32
            k4  = ( c1 * k4 ) & 0xFFFFFFFF
            h4 ^= k4

            h4 = ( h4 << 13 | h4 >> 19 ) & 0xFFFFFFFF # inlined ROTL32
            h4 = ( h1 + h4 ) & 0xFFFFFFFF
            h4 = ( h4 * 5 + 0x32ac3b17 ) & 0xFFFFFFFF

        #tail
        tail_index = nblocks * 16
        k1 = 0
        k2 = 0
        k3 = 0
        k4 = 0
        tail_size = length & 15

        if tail_size >= 15:
            k4 ^= key[ tail_index + 14 ] << 16
        if tail_size >= 14:
            k4 ^= key[ tail_index + 13 ] << 8
        if tail_size >= 13:
            k4 ^= key[ tail_index + 12 ]

        if tail_size > 12:
            k4  = ( k4 * c4 ) & 0xFFFFFFFF
            k4  = ( k4 << 18 | k4 >> 14 ) & 0xFFFFFFFF # inlined ROTL32
            k4  = ( k4 * c1 ) & 0xFFFFFFFF
            h4 ^= k4

        if tail_size >= 12:
            k3 ^= key[ tail_index + 11 ] << 24
        if tail_size >= 11:
            k3 ^= key[ tail_index + 10 ] << 16
        if tail_size >= 10:
            k3 ^= key[ tail_index +  9 ] << 8
        if tail_size >=  9:
            k3 ^= key[ tail_index +  8 ]

        if tail_size > 8:
            k3  = ( k3 * c3 ) & 0xFFFFFFFF
            k3  = ( k3 << 17 | k3 >> 15 ) & 0xFFFFFFFF # inlined ROTL32
            k3  = ( k3 * c4 ) & 0xFFFFFFFF
            h3 ^= k3

        if tail_size >= 8:
            k2 ^= key[ tail_index + 7 ] << 24
        if tail_size >= 7:
            k2 ^= key[ tail_index + 6 ] << 16
        if tail_size >= 6:
            k2 ^= key[ tail_index + 5 ] << 8
        if tail_size >= 5:
            k2 ^= key[ tail_index + 4 ]

        if tail_size > 4:
            k2  = ( k2 * c2 ) & 0xFFFFFFFF
            k2  = ( k2 << 16 | k2 >> 16 ) & 0xFFFFFFFF # inlined ROTL32
            k2  = ( k2 * c3 ) & 0xFFFFFFFF
            h2 ^= k2

        if tail_size >= 4:
            k1 ^= key[ tail_index + 3 ] << 24
        if tail_size >= 3:
            k1 ^= key[ tail_index + 2 ] << 16
        if tail_size >= 2:
            k1 ^= key[ tail_index + 1 ] << 8
        if tail_size >= 1:
            k1 ^= key[ tail_index + 0 ]

        if tail_size > 0:
            k1  = ( k1 * c1 ) & 0xFFFFFFFF
            k1  = ( k1 << 15 | k1 >> 17 ) & 0xFFFFFFFF # inlined ROTL32
            k1  = ( k1 * c2 ) & 0xFFFFFFFF
            h1 ^= k1

        #finalization
        h1 ^= length
        h2 ^= length
        h3 ^= length
        h4 ^= length

        h1 = ( h1 + h2 ) & 0xFFFFFFFF
        h1 = ( h1 + h3 ) & 0xFFFFFFFF
        h1 = ( h1 + h4 ) & 0xFFFFFFFF
        h2 = ( h1 + h2 ) & 0xFFFFFFFF
        h3 = ( h1 + h3 ) & 0xFFFFFFFF
        h4 = ( h1 + h4 ) & 0xFFFFFFFF

        h1 = fmix( h1 )
        h2 = fmix( h2 )
        h3 = fmix( h3 )
        h4 = fmix( h4 )

        h1 = ( h1 + h2 ) & 0xFFFFFFFF
        h1 = ( h1 + h3 ) & 0xFFFFFFFF
        h1 = ( h1 + h4 ) & 0xFFFFFFFF
        h2 = ( h1 + h2 ) & 0xFFFFFFFF
        h3 = ( h1 + h3 ) & 0xFFFFFFFF
        h4 = ( h1 + h4 ) & 0xFFFFFFFF

        return ( h4 << 96 | h3 << 64 | h2 << 32 | h1 )

    key = bytearray( xencode(key) )

    if x64arch:
        return hash128_x64( key, seed )
    else:
        return hash128_x86( key, seed )


def hash64( key, seed = 0x0, x64arch = True ):
    ''' Implements 64bit murmur3 hash. Returns a tuple. '''

    hash_128 = hash128( key, seed, x64arch )

    unsigned_val1 = hash_128 & 0xFFFFFFFFFFFFFFFF
    if unsigned_val1 & 0x8000000000000000 == 0:
        signed_val1 = unsigned_val1
    else:
        signed_val1 = -( (unsigned_val1 ^ 0xFFFFFFFFFFFFFFFF) + 1 )

    unsigned_val2 = ( hash_128 >> 64 ) & 0xFFFFFFFFFFFFFFFF
    if unsigned_val2 & 0x8000000000000000 == 0:
        signed_val2 = unsigned_val2
    else:
        signed_val2 = -( (unsigned_val2 ^ 0xFFFFFFFFFFFFFFFF) + 1 )

    return ( int( signed_val1 ), int( signed_val2 ) )


def hash_bytes( key, seed = 0x0, x64arch = True ):
    ''' Implements 128bit murmur3 hash. Returns a byte string. '''

    hash_128 = hash128( key, seed, x64arch )

    bytestring = ''

    for i in xrange(0, 16, 1):
        lsbyte = hash_128 & 0xFF
        bytestring = bytestring + str( chr( lsbyte ) )
        hash_128 = hash_128 >> 8

    return bytestring


if __name__ == "__main__":
    import argparse
    
    parser = argparse.ArgumentParser( 'pymurmur3', 'pymurmur [options] "string to hash"' )
    parser.add_argument( '--seed', type = int, default = 0 )
    parser.add_argument( 'strings', default = [], nargs='+')
    
    opts = parser.parse_args()
    
    for str_to_hash in opts.strings:
        sys.stdout.write( '"%s" = 0x%08X\n' % ( str_to_hash, hash( str_to_hash ) ) )



import os
import json
import argparse

from waitress import serve

from mcrit.server.wsgi import app
from mcrit.client.McritClient import McritClient
from mcrit.Worker import Worker


def runWorker(profiling=False):
    worker = Worker(profiling=profiling)
    worker.run()


def runServer(profiling=False):
    wrapped_app = app
    if profiling:
        print("[!] Running as profiled application.")
        from werkzeug.middleware.profiler import ProfilerMiddleware
        profile_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "profiler")
        os.makedirs(profile_dir, exist_ok=True)
        wrapped_app = ProfilerMiddleware(
            wrapped_app,
            restrictions=[30],
            profile_dir=profile_dir,
            filename_format="{method}-{path}-{time:.0f}-{elapsed:.0f}ms.prof",
        )
    # TODO consider allowing an argument to pass an configuration for initial setup of the instance
    serve(wrapped_app, listen="*:8000")


def runClient(arguments):
    client = McritClient()
    result = None
    if ARGS.client_command == "status":
        result = client.getStatus()
        print(result)
    elif ARGS.client_command == "import":
        if os.path.isfile(ARGS.filepath):
            with open(ARGS.filepath) as fin:
                result = client.addImportData(json.load(fin))
                print(result)
    elif ARGS.client_command == "export":
        sample_ids = None
        if ARGS.sample_ids is not None:
            sample_ids = [int(i) for i in ARGS.sample_ids.split(",")]
        result = client.getExportData(sample_ids=sample_ids)
        with open(ARGS.filepath, "w") as fout:
            json.dump(result, fout, indent=1)
        print(f"wrote export to {ARGS.filepath}.")
    elif ARGS.client_command == "search":
        result = client.search(ARGS.search_term)
        print(result)
    elif ARGS.client_command == "queue":
        result = client.getQueueData(filter=ARGS.filter)
        for entry in result:
            job_id = entry.job_id
            result_id = entry.result
            created = entry.created_at if entry.created_at is not None else "-"
            started = entry.started_at if entry.started_at is not None else "-"
            finished = entry.finished_at if entry.finished_at is not None else "-"
            method = entry.parameters
            progress = entry.progress
            print(f"{job_id} {result_id} | {created} {started} {finished} | {method} - {progress}")


parser = argparse.ArgumentParser()
subparsers = parser.add_subparsers(dest="command")

parser_server = subparsers.add_parser("server")
parser_server.add_argument("--profile", help="Profile server. Requires werkzeug package.", action="store_true")
parser_worker = subparsers.add_parser("worker")
parser_worker.add_argument("--profile", help="Profile worker. Requires cProfile.", action="store_true")
# create a set of subparsers for all client commands
parser_client = subparsers.add_parser("client")
subparser_client = parser_client.add_subparsers(dest="client_command")
client_status = subparser_client.add_parser("status")
client_import = subparser_client.add_parser("import")
client_import.add_argument("filepath", type=str, help="Import a given <filepath> containing MCRIT data into the storage.")
client_export = subparser_client.add_parser("export")
client_export.add_argument("filepath", type=str, help="Export the full storage into <filepath>.")
client_export.add_argument("--sample_ids", type=str, help="Limit export to a list of comma-separated <sample_ids>.")
client_search = subparser_client.add_parser("search")
client_search.add_argument("search_term", type=str, help="Conduct a search with a given <search_term>.")
client_queue = subparser_client.add_parser("queue")
client_queue.add_argument("--filter", type=str, help="Filter queue entries with this term.")

ARGS = parser.parse_args()

if ARGS.command == "server":
    runServer(profiling = ARGS.profile)
elif ARGS.command == "worker":
    runWorker(profiling = ARGS.profile)
elif ARGS.command == "client":
    runClient(ARGS)
else:
    parser.print_usage()



from datetime import datetime
import hashlib
import json
import os
from sqlite3 import Timestamp
import time
from functools import wraps
import logging


# Only do basicConfig if no handlers have been configured
if len(logging._handlerList) == 0:
    logging.basicConfig(level=logging.INFO, format="%(asctime)-15s %(message)s")
LOGGER = logging.getLogger(__name__)



################# Caller ##################
class BaseRemoteCallerClass:
    def __init__(self, queue):
        self.queue = queue

    #### JOB CONTROL ####
    def JobStatus(self, job_id):
        # returns status, msg
        # status can be "NotFound", "Running", "Done", "Error"
        # msg can be the error message or Result_id in case of Done
        # Check mongoqueue for job_id if Error or Running -> Return
        # Check results-collection/job-archive for results with original_job_id=job_id
        # -> either return "Done" with result / where to find result, OR NotFound
        pass

    def getQueueData(self, filter=None):
        LOGGER.debug(f"getQueueData(filter={filter}):")
        if filter is not None:
            # TODO apply filter to more fields
            return [job._data for job in self.queue.get_jobs() if filter in job.parameters]
        return [job._data for job in self.queue.get_jobs()]

    def getJob(self, job_id):
        LOGGER.debug("GetJob: %s", job_id)
        return self.queue.get_job(job_id)

    def getJobData(self, job_id):
        LOGGER.debug("GetJobData: %s", job_id)
        return self.queue.get_job(job_id)._data

    def getResultForJob(self, job_id):
        result = None
        res_id = self.queue.get_job(job_id).result
        if res_id is not None:
            LOGGER.debug("GetResultForJob: %s -> %s", job_id, res_id)
            result = self.queue._grid_to_dicts(res_id)
        return result

    def getJobIdForResult(self, result_id):
        job_id = None
        meta = self.queue._grid_to_meta(result_id)
        if meta is not None and "job" in meta:
            job_id = meta["job"]
            LOGGER.debug("getJobIdForResult: %s -> %s", result_id, job_id)
        return job_id

    def getResult(self, res_id):
        result_dict = self.queue._grid_to_dicts(res_id)
        return result_dict

    def awaitResult(self, job_id):
        job = self.queue.get_job(job_id)
        terminated = job._is_terminated(use_cached=True)
        failed = job.is_failed
        result_id = job.result
        while (result_id is None) and (not terminated) and (not failed):
            time.sleep(0.05)
            # TODO what if job is already killed by clear?
            job = self.queue.get_job(job_id)
            terminated = job._is_terminated(use_cached=True)
            failed = job.is_failed
            result_id = job.result
        return result_id


########### START Class Metaprogramming
# Add proxies to clsCallee to clsCaller
# clsCallee is expected to be derived from QueueRemoteCallee
# clsCaller needs an self.queue attribute
def addRemoteCallFunctions(clsCallee, clsCaller):
    method_list = [func for func in dir(clsCallee) if callable(getattr(clsCallee, func)) and not func.startswith("__")]
    for name, method in zip(method_list, [getattr(clsCallee, func) for func in method_list]):
        if hasattr(method, "remote") and method.remote == True:
            new_method = RemotifyFunctionWrapper(method)
            new_method.__qualname__ = ".".join([clsCaller.__qualname__, new_method.__name__])
            setattr(clsCaller, name, new_method)
    return clsCaller


# Usage:
# MyCallerClass = QueueRemoteCallee(MyWorkerClass)
# or
# class MyAdvancedCallerClass(QueueRemoteCallee(MyWorkerClass)):
#     ...
def QueueRemoteCaller(clsCallee):
    class RemoteCallerClass(BaseRemoteCallerClass):
        pass

    return addRemoteCallFunctions(clsCallee, RemoteCallerClass)


########### END Class Metaprogramming

# Wrapper that creates a remote call proxy for a given method
def RemotifyFunctionWrapper(function):
    def submitPayloadQueue(self, payload, await_jobs):
        return str(self.queue.put(payload, await_jobs=await_jobs))

    # remote call proxy
    def remote_call_function(self, *params, await_jobs=None, force_recalculation=False, **kwparams):
        name = function.__name__

        # join file locations:
        file_locations = function.kwfile_locations + function.file_locations
        json_locations = function.kwjson_locations + function.json_locations

        # get rearranged parameters
        params, file_params = rearrange_params(params, kwparams, file_locations, json_locations)

        # get descriptor:
        hashes = hash_all(file_params)
        descriptor = get_descriptor(name, params, hashes)

        # Evaluate Cached jobs
        if not force_recalculation:
            cachedJobId = self.queue.get_cached_job_id({"descriptor": descriptor})
            if cachedJobId is not None:
                return str(cachedJobId)

        # Upload
        grid_params = upload_file_params(self, file_params, hashes)

        # Submit
        payload = _createJobPayload(name, params, grid_params, descriptor)
        if await_jobs is None:
            await_jobs = []
        job_id = submitPayloadQueue(self, payload, await_jobs)

        # Add job ids to parameters
        add_job_id_to_files(self, job_id, grid_params)

        return job_id

    return wraps(function)(remote_call_function)


### Helper functions
def sha256(data):
    return hashlib.sha256(data).hexdigest()


def to_binary(dicts):
    return json.dumps(dicts, sort_keys=True).encode("ascii")


## helpers for rearange_params
def add_list_to_dict(Dict, List):
    for i, entry in enumerate(List):
        Dict[i] = entry


def stringify_keys(d):
    return {str(key): val for key, val in d.items()}


def json_param_preprocessing(params, locations):
    for f in locations:
        if f in params:
            params[f] = to_binary(params[f])
    return params


def file_param_split(params, locations):
    files = {}
    for f in locations:
        if f in params:
            files[str(f)] = params[f]
            params[f] = None
    return params, files


# This function merges params and kwparams, by using VALUE-KEYS like "0", "1", ...
# Then the parameters are split again, into "normal" params, and file_params, which are transmitted via gridfs
def rearrange_params(params, kwparams, file_locations, json_locations):
    all_file_locations = file_locations + json_locations
    # join parameters:
    add_list_to_dict(kwparams, params)
    # all parameters are in params now:
    params = kwparams
    # do json preprocessing, replacing datastructures with strings
    params = json_param_preprocessing(params, json_locations)
    # split params into params and file_params
    params, file_params = file_param_split(params, all_file_locations)
    # stringify keys for params
    params = stringify_keys(params)
    # file_params = stringify_keys(file_params) #already done bei file_param_split
    return params, file_params


def hash_all(d):
    return {key: sha256(val) for key, val in d.items()}


def get_descriptor(name, params, hashes):
    return json.dumps((name, params, hashes), sort_keys=True)


def upload_file_params(self, file_params, hashes):
    grid_params = {}
    for key, val in file_params.items():
        grid = self.queue.get_file_by_hash_inc_lock(hashes[key])
        if grid is None:
            metadata = {"sha256": hashes[key], "tmp_lock": 1, "jobs": []}
            grid = self.queue._file_to_grid(val, metadata=metadata)
        grid_params[key] = grid
    return grid_params


def add_job_id_to_files(self, job_id, grid_params):
    for key, val in grid_params.items():
        self.queue.add_job_id_to_file(job_id, val)


# Function to encode call
def _createJobPayload(method_name, params, grid_params, descriptor):
    payload = {
        "method": method_name,
        "params": json.dumps(params),
        "file_params": json.dumps(grid_params),
        "descriptor": descriptor,
    }
    return payload



################# Callee ##################

# Marks Functions within a QueueRemoteCallee
def Remote(progress=False, file_locations=[], kwfile_locations=[], json_locations=[], kwjson_locations=[]):
    def change_function(function):
        function.remote = True
        function.progressor = progress
        function.file_locations = file_locations
        function.kwfile_locations = kwfile_locations
        function.json_locations = json_locations
        function.kwjson_locations = kwjson_locations
        return function

    return change_function


# Class to be extended
# It is derived from BaseRemoteCallerClass -> it can also query job results, etc
class QueueRemoteCallee(BaseRemoteCallerClass):
    def __init__(self, queue, profiling_path=None):
        self.queue = queue
        self.queue.clean()
        self._alive = True
        self.t_last_cleanup = time.time()
        if profiling_path is not None:
            self._executeJob = self.profiling_wrapper(self._executeJob, profiling_path)
    
    def profiling_wrapper(self, function, profiling_path):
        import cProfile
        @wraps(function)
        def wrapped_function(job, *args, **kwargs):
            start = datetime.utcnow()
            with cProfile.Profile() as pr:
                result = function(job, *args, **kwargs)
            end = datetime.utcnow()
            method = job.payload["method"]
            duration = int((end - start).total_seconds()*1000)
            filename = f"WORKER-{method}-{int(start.timestamp())}-{duration}ms.prof"
            pr.dump_stats(os.path.join(profiling_path, filename))
            return result
        return wrapped_function

    def _receive_files(self, d):
        for key, val in d.items():
            d[key] = self.queue._grid_to_file(val, results_only=False)
        return d

    def _decodeJobPayload(self, job_payload):
        payload = job_payload
        method_name = job_payload["method"]
        method = getattr(self, method_name)
        if not (hasattr(method, "remote") and method.remote == True):
            raise NotImplementedError

        params = json.loads(job_payload["params"])
        params = restore_int_keys(params)

        file_params = json.loads(job_payload["file_params"])
        file_params = restore_int_keys(file_params)

        file_params = self._receive_files(file_params)
        file_params = json_post_processing(file_params, method.json_locations + method.kwjson_locations)

        params = join_files_to_params(params, file_params)
        params, kwparams = split_list_dict(params)

        return method, params, kwparams

    def _executeJobPayload(self, job_payload, job):
        LOGGER.debug("DECODE JOB: %s", job_payload)
        method, params, kwparams = self._decodeJobPayload(job_payload)
        # Add progress reporter if necessary:
        if method.progressor:
            LOGGER.debug("kwparams: %s", kwparams)
            kwparams["progress_reporter"] = JobProgressReporter(job, 0.1)
        LOGGER.debug("EXECUTE JOB: %s", job_payload)
        result = method(*params, **kwparams)
        LOGGER.debug("FINISHED JOB: %s", job_payload)
        return result

    def _executeJob(self, job):
        if time.time() - self.t_last_cleanup >= self.queue.clean_interval:
            self.queue.clean()
            self.t_last_cleanup = time.time()
        try:
            with job as j:
                LOGGER.info("Processing Remote Job: %s", job)
                result = self._executeJobPayload(j["payload"], job)
                LOGGER.debug("Remote Job Result: %s", result)
                # ensure we always have a job_id for finished job payloads
                job.result = self.queue._dicts_to_grid(result, metadata={"result": True, "job": job.job_id})
                LOGGER.info("Finished Remote Job: %s", job)
        except Exception as exc:
            pass

    def run(self):
        self._alive = True
        while self._alive:
            job = self.queue.next()
            if job:
                LOGGER.debug("Found job")
                self._executeJob(job)
            else:
                time.sleep(0.1)

    def terminate(self):
        self._alive = False


### Helper functions


def from_binary(binary):
    return json.loads(binary.decode("ascii"))


def join_files_to_params(params, file_params):
    for key, val in file_params.items():
        params[key] = val
    return params


def split_list_dict(in_dict):
    L = []
    i = 0
    while i in in_dict:
        L.append(in_dict[i])
        i += 1
    out_dict = {key: val for key, val in in_dict.items() if type(key) != int}
    return L, out_dict


def json_post_processing(params, locations):
    for f in locations:
        if f in params:
            params[f] = from_binary(params[f])
    return params


def restore_int_keys(params):
    int_dict = {}
    del_keys = []
    for key, val in params.items():
        try:
            key_int = int(key)
            int_dict[key_int] = val
            del_keys.append(key)
        except:
            pass

    for k in del_keys:
        del params[k]
    params.update(int_dict)
    return params

################################### progress iterator


class NoProgressReporter:
    def set_total(self, total):
        pass

    def step(self):
        pass


class JobProgressReporter:
    def __init__(self, job, report_interval):
        self._job = job
        self._total = 1
        self._last_update = time.time()
        self._report_interval = report_interval
        self._count = 0
        self._submit_progress(0)

    # Total should be set after creation of this object, in the called function.
    # If total is not set, progress will be reported as absolute number of steps.
    def set_total(self, total):
        self._total = total

    def step(self):
        self._count += 1
        now = time.time()
        if now - self._last_update > self._report_interval:
            progress = self._count / self._total
            self._submit_progress(progress)
            if self._check_terminated():
                self._terminate_job()
            self._last_update = now

    def _submit_progress(self, progress):
        self._job.progressor(count=progress)

    def _check_terminated(self):
        return self._job._is_terminated()

    def _terminate_job(self):
        raise Exception("The job received a terminatation message while reporting progress")


class EmptyProgressWrapper:
    def __init__(self, iterable, total=None):
        self._iterable = iterable
        self._iterator = None
        if total is not None:
            self._length = total
        else:
            self._length = len(iterable)
        self._progress_reporter = JobProgressReporter(None, 1)

    def __iter__(self):
        self._iterator = iter(self._iterable)
        return self

    def __next__(self):
        self._progress_reporter.step()
        return next(self._iterator)



import logging
from re import M

import falcon

from mcrit.server.utils import timing, jsonify
from mcrit.index.MinHashIndex import MinHashIndex

LOGGER = logging.getLogger(__name__)


class FunctionResource:
    def __init__(self, index: MinHashIndex):
        self.index = index

    @timing
    def on_get(self, req, resp, function_id=None):
        LOGGER.info("FunctionResource.on_get_function")
        query_with_xcfg = False 
        if "with_xcfg" in req.params:
            query_with_xcfg = req.params["with_xcfg"].lower().strip() == "true"
        if not self.index.isFunctionId(function_id):
            resp.data = jsonify(
                {
                    "status": "failed",
                    "data": {"message": "We don't have a function with that id."},
                }
            )
            resp.status = falcon.HTTP_404
            return
        data = self.index.getFunctionById(function_id, with_xcfg=query_with_xcfg).toDict()
        resp.data = jsonify(
            {
                "status": "successful",
                "data": data,
            }
        )

    @timing
    def on_get_collection(self, req, resp):
        LOGGER.info("FunctionResource.on_get_collection")
        # parse optional request parameters
        start_index = 0 
        if "start" in req.params:
            try:
                start_index = int(req.params["start"])
            except:
                pass
        limit_function_count = 0 
        if "limit" in req.params:
            try:
                limit_function_count = int(req.params["limit"])
            except:
                pass
        function_overview = {}
        function_entries = self.index.getFunctions(start_index, limit_function_count)
        for function_entry in function_entries:
            function_overview[function_entry.function_id] = function_entry.toDict()
        resp.data = jsonify({"status": "successful", "data": function_overview})



#!/usr/bin/env python3

import logging
import os
import random
import sys

logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)


class ShingleLoader:

    WEIGHT_STRATEGY_ALL_SHINGLERS_EQUAL = 1
    WEIGHT_STRATEGY_SHINGLER_WEIGHTS = 2

    def __init__(self, shingler_config):
        self._config = shingler_config
        self._python_files = self._getPythonFiles()
        self._shingler_classes = self._getShinglerClasses()
        self._updateXorValues()

    def _updateXorValues(self):
        if not self._config.SHINGLERS_XOR_VALUES:
            random.seed(self._config.SHINGLERS_SEED)
            self._config.SHINGLERS_XOR_VALUES = [0] + [
                random.randint(0, 0xFFFFFFFF) for _ in range(max(self._config.SHINGLERS_WEIGHTS.values()))
            ]

    def getShinglers(self):
        """Get a (name-)sorted list of shinglers as specified in the config"""
        if self._config.SHINGLER_WEIGHT_STRATEGY == self.WEIGHT_STRATEGY_SHINGLER_WEIGHTS:
            return self._initWeightedShinglers()
        elif self._config.SHINGLER_WEIGHT_STRATEGY == self.WEIGHT_STRATEGY_ALL_SHINGLERS_EQUAL:
            return self._initAllShinglers()
        raise NotImplementedError("Unknown Shingler Weight Strategy.")

    def _initWeightedShinglers(self):
        shingler_instances = []
        for shingler_cls in self._shingler_classes:
            shingler_name = shingler_cls(self._config).getName()
            if shingler_name in self._config.SHINGLERS_WEIGHTS and self._config.SHINGLERS_WEIGHTS[shingler_name]:
                weight = self._config.SHINGLERS_WEIGHTS[shingler_name]
                shingler = shingler_cls(self._config, weight)
                shingler_instances.append(shingler)
        LOGGER.info(
            "Using Shingler Setup: %s",
            ", ".join(
                sorted(["%s (%d)" % (shingler.getName(), shingler.getWeight()) for shingler in shingler_instances])
            ),
        )
        return sorted(shingler_instances)

    def _initAllShinglers(self):
        return sorted([shingler_cls(self._config) for shingler_cls in self._shingler_classes])

    def _getPythonFiles(self):
        python_files = [
            os.path.join(self._config.SHINGLER_DIR, candidate_file)
            for candidate_file in os.listdir(self._config.SHINGLER_DIR)
            if (
                os.path.isfile(os.path.join(self._config.SHINGLER_DIR, candidate_file))
                and candidate_file.endswith("Shingler.py")
                and candidate_file != "AbstractShingler.py"
            )
        ]
        return python_files

    def _getShinglerClasses(self):
        classes = []
        for python_file in sorted(self._python_files):
            file_path = os.path.split(os.path.abspath(python_file))[0]
            if file_path not in sys.path:
                sys.path.append(file_path)
            python_module = (os.path.split(python_file)[1])[:-3]
            LOGGER.info("loading shingler module: %s", python_module)
            module = __import__(python_module, fromlist=[python_module])
            module_class = getattr(module, python_module)
            classes.append(module_class)
        return classes



import re
import logging

from mcrit.server.utils import timing, jsonify
from mcrit.index.MinHashIndex import MinHashIndex


LOGGER = logging.getLogger(__name__)


class BlocksResource:
    def __init__(self, index: MinHashIndex):
        self.index = index

    @timing
    def on_get_unique_blocks_for_family(self, req, resp, family_id:int):
        LOGGER.info("BlocksResource.on_get_unique_blocks_for_family")
        blocks_result = {}
        samples = self.index.getSamplesByFamilyId(family_id)
        target_sample_ids = [sample.sample_id for sample in samples]
        blocks_result = self.index.getUniqueBlocks(target_sample_ids, family_id=family_id)
        resp.data = jsonify({"status": "successful", "data": blocks_result})

    @timing
    def on_get_unique_blocks_for_samples(self, req, resp, comma_separated_sample_ids=None):
        LOGGER.info("BlocksResource.on_get_unique_blocks")
        blocks_result = {}
        if re.match("^\d+(?:[\s]*,[\s]*\d+)*$", comma_separated_sample_ids):
            target_sample_ids = [int(sample_id) for sample_id in comma_separated_sample_ids.split(",")]
            blocks_result = self.index.getUniqueBlocks(target_sample_ids)
        resp.data = jsonify({"status": "successful", "data": blocks_result})



from typing import List, Literal, Union


##### Defining all Node types #####
class Node:
    pass

NodeType = Union["SearchTermNode", "SearchConditionNode", "NotNode", "AndNode", "OrNode"]

class _ListNode(Node):
    children: List[NodeType]
    _name = ""

    def __init__(self, children:List[NodeType], flatten=True):
        self.children = children
        if flatten:
            self.flatten()

    def flatten(self):
        flattened_children = []
        for child in self.children:
            if isinstance(child, type(self)):
                flattened_children.extend(child.children)
            else:
                flattened_children.append(child)
        self.children = flattened_children

    def __repr__(self) -> str:
        children_str = ", ".join([str(child) for child in self.children])
        return f"{self._name}({children_str})"

class AndNode(_ListNode):
    _name = "And"

class OrNode(_ListNode):
    _name = "Or"

class NotNode(Node):
    child: NodeType 

    def __init__(self, child:NodeType):
        self.child = child

    def __repr__(self) -> str:
        return f"Not({str(self.child)})"

class SearchTermNode(Node):
    value: str

    def __init__(self, value):
        self.value = value

    def __repr__(self) -> str:
        return f"Search('{self.value}')"

class SearchConditionNode(Node):
    field: str
    operator: Literal["", "=", "<", "<=", ">", ">=", "!=", "?", "!?"]
    value: str

    def __init__(self, field, operator, value):
        self.field = field 
        self.operator = operator 
        self.value = value

    def __repr__(self) -> str:
        if self.operator != "":
            operator_str = self.operator
        else:
            operator_str = ":"
        return f"Condition('{self.field}' {operator_str} '{self.value}')"


#### Code operating on trees #####

class BaseVisitor:
    """
    BaseVisitor.visit traverses a given tree and returns a newly build identical tree.
    BaseVisitor is used as base class for more advanced tree transformations.
    """
    def visit(self, node:NodeType):
        if isinstance(node, SearchTermNode):
            return self.visitSearchTermNode(node)
        if isinstance(node, SearchConditionNode):
            return self.visitSearchConditionNode(node)
        if isinstance(node, AndNode):
            return self.visitAndNode(node)
        if isinstance(node, OrNode):
            return self.visitOrNode(node)
        if isinstance(node, NotNode):
            return self.visitNotNode(node)

    def visitOrNode(self, node:OrNode):
        visited_children = [self.visit(child) for child in node.children]
        return OrNode(visited_children)

    def visitAndNode(self, node:AndNode):
        visited_children = [self.visit(child) for child in node.children]
        return AndNode(visited_children)

    def visitNotNode(self, node:NotNode):
        return NotNode(self.visit(node.child))

    def visitSearchTermNode(self, node:SearchTermNode):
        return node

    def visitSearchConditionNode(self, node:SearchConditionNode):
        return node


class SearchFieldResolver(BaseVisitor):
    """
    SearchFieldResolver replaces "complex" SearchTermNodes with more simple SearchConditionNodes.
    This adds context information (i.e. in which fields do we want to search?).
    """
    def __init__(self, search_fields, conditional_search_fields = None) -> None:
        """
        search_fields: list of fields to search in
        conditional_search_fields: tuple (search_field, condition).
                                   Only search for node.value in search_field if condition(node.value) == True
        """
        super().__init__()
        self.search_fields = search_fields
        if conditional_search_fields is None:
            conditional_search_fields = []
        self.conditional_search_fields = conditional_search_fields

    def visitSearchTermNode(self, node: SearchTermNode):
        children = []
        for field in self.search_fields:
            children.append(
                SearchConditionNode(field, "?", node.value)
            )
        for field, condition in self.conditional_search_fields:
            if condition(node.value):
                children.append(
                    SearchConditionNode(field, "?", node.value)
                )
        return OrNode(children)


class FilterSingleElementLists(BaseVisitor):
    """
    FilterSingleElementLists replaces And(node) as well as Or(node) by node.
    """
    def visitAndNode(self, node):
        children = [self.visit(child) for child in node.children]
        if len(children) == 1:
            return children[0]
        return AndNode(children)

    def visitOrNode(self, node):
        children = [self.visit(child) for child in node.children]
        if len(children) == 1:
            return children[0]
        return OrNode(children)

class Negate(BaseVisitor):
    """
    Negate negates a tree and uses De Morgans Law to move Not Nodes down the tree.
    Requires the tree to be free from SearchTermNodes (see SearchFieldResolver).
    The ouput is free from Not nodes.
    """
    def visitSearchTermNode(self, node: SearchTermNode):
        raise NotImplementedError
    
    def visitSearchConditionNode(self, node: SearchConditionNode):
        reverse_op_dict = {
            "": "!=",
            "=": "!=",
            "<": ">=",
            "<=": ">",
            ">": "<=",
            ">=": "<",
            "!=": "=",
            "?": "!?",
            "!?": "?",
        }
        return SearchConditionNode(node.field, reverse_op_dict[node.operator], node.value)

    def visitOrNode(self, node:OrNode):
        visited_children = [self.visit(child) for child in node.children]
        return AndNode(visited_children)

    def visitAndNode(self, node:AndNode):
        visited_children = [self.visit(child) for child in node.children]
        return OrNode(visited_children)

    def visitNotNode(self, node:NotNode):
        return PropagateNot().visit(node.child)
        

class PropagateNot(BaseVisitor):
    """
    PropagateNot uses De Morgans Law to move Not Nodes down the tree.
    Requires the tree to be free from SearchTermNodes (see SearchFieldResolver).
    The ouput is free from Not nodes.
    """
    def visitNotNode(self, node:NotNode):
        return Negate().visit(node.child)



import logging

import falcon

from mcrit.server.utils import timing, jsonify, getMatchingParams
from mcrit.index.MinHashIndex import MinHashIndex


LOGGER = logging.getLogger(__name__)


class MatchResource:
    def __init__(self, index: MinHashIndex):
        self.index = index

    @timing
    def on_get_sample(self, req, resp, sample_id=None):
        LOGGER.info("SampleResource.on_get_matches")
        parameters = getMatchingParams(req.params)
        if not self.index.isSampleId(sample_id):
            resp.data = jsonify(
                {
                    "status": "failed",
                    "data": {"message": "We don't have a sample with that id."},
                }
            )
            resp.status = falcon.HTTP_404
            return

        sample_matches = self.index.getMatchesForSample(sample_id, **parameters)
        resp.data = jsonify({"status": "successful", "data": sample_matches})
        # resp.status = falcon.HTTP_202
        # resp.location = "/matches/123"
        # resp.retry_after = 2

    @timing
    def on_get_sample_cross(self, req, resp, sample_ids=None):
        LOGGER.info("SampleResource.on_get_matches_cross")
        parameters = getMatchingParams(req.params)
        sample_ids_list = [int(id) for id in sample_ids.split(",")]
        cross_matches = self.index.getMatchesCross(sample_ids_list, **parameters)
        resp.data = jsonify({"status": "successful", "data": cross_matches})

    @timing
    def on_get_sample_vs(self, req, resp, sample_id=None, sample_id_b=None):
        # NOTE: We don't need to check if the kw parameters are None. The routing ensures that they are always set.
        LOGGER.info("SampleResource.on_get_matches_vs")
        parameters = getMatchingParams(req.params)
        if not self.index.isSampleId(sample_id) or not self.index.isSampleId(sample_id_b):
            resp.data = jsonify(
                {
                    "status": "failed",
                    "data": {"message": "We don't have a sample with that id."},
                }
            )
            resp.status = falcon.HTTP_404
            return
        sample_matches = self.index.getMatchesForSampleVs(sample_id, sample_id_b, **parameters)
        resp.data = jsonify({"status": "successful", "data": sample_matches})

    @timing
    def on_get_function(self, req, resp, function_id=None):
        resp.status = falcon.HTTP_NOT_IMPLEMENTED
        return

    @timing
    def on_get_function_vs(self, req, resp, function_id=None, function_id_b=None):
        LOGGER.info("SampleResource.on_get_function_vs")
        if not self.index.isFunctionId(function_id) or not self.index.isFunctionId(function_id_b):
            resp.data = jsonify(
                {
                    "status": "failed",
                    "data": {"message": "We don't have a sample with that id."},
                }
            )
            resp.status = falcon.HTTP_404
            return
        function_match = self.index.getMatchesFunctionVs(function_id, function_id_b)
        resp.data = jsonify({"status": "successful", "data": function_match})



# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

with open("README.md") as f:
    README = f.read()

requirements = [
    "dataclasses",
    "falcon>=2.0.0",
    "fastcluster",
    "mmh3>=2.5.1",
    "numpy",
    "picblocks>=1.1.2",
    "pymongo",
    "pyparsing>=3",
    "requests",
    "scipy",
    "smda>=1.3.0",
    "tqdm",
    "waitress",
]

setup(
    name='mcrit',
    version="0.19.4",
    description='MCRIT is a framework created for simplified application of the MinHash algorithm to code similarity.',
    long_description_content_type="text/markdown",
    long_description=README,
    author='Daniel Plohmann, Steffen Enders, Paul Hordiienko, Manuel Blatt',
    author_email='daniel.plohmann@fkie.fraunhofer.de',
    url='https://github.com/danielplohmann/mcrit',
    license="NU General Public License v3 (GPLv3)",
    packages=find_packages(exclude=("tests", "data", "docs", "examples", "plugins")),
    install_requires=requirements,
    data_files=[
        ("", ["LICENSE"]),
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Security",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
)



from random import sample
from typing import TYPE_CHECKING, Dict, List, Optional

from mcrit.storage.SampleEntry import SampleEntry
from mcrit.storage.MatchedSampleEntry import MatchedSampleEntry
from mcrit.storage.MatchedFunctionEntry import MatchedFunctionEntry
import mcrit.matchers.MatcherInterface as MatcherInterface

if TYPE_CHECKING:  # pragma: no cover
    from mcrit.storage.SampleEntry import SampleEntry

# Dataclass, post init
# constructor -> .fromSmdaFunction
# assume sample_entry, smda_function always available

class MatchingResult(object):
    reference_sample_entry: "SampleEntry"
    other_sample_entry: "SampleEntry"
    match_aggregation: Dict
    sample_matches: List["MatchedSampleEntry"]
    function_matches: List["MatchedFunctionEntry"]
    # function_id -> [(family_id, sample_id), ...]
    library_matches: Dict
    unique_family_scores_per_sample: Dict
    is_family_filtered: bool
    is_sample_filtered: bool
    is_function_filtered: bool
    is_score_filtered: bool
    is_sample_count_filtered: bool
    is_family_count_filtered: bool

    def __init__(self, sample_entry: "SampleEntry") -> None:
        self.reference_sample_entry = sample_entry

    def getFamilyNameByFamilyId(self, family_id):
        family_name = ""
        for sample_match in self.sample_matches:
            if sample_match.family_id == family_id:
                family_name = sample_match.family
                break
        return family_name

    def getFilteredSampleMatch(self):
        if not self.is_sample_filtered or not len(self.sample_matches):
            return None
        else:
            return self.sample_matches[0]

    def excludeLibraryMatches(self):
        """ reduce contained matches to those where none of the matches is with a library (transitive library identification) """
        library_matches = [matches for matches in self.library_matches.values() if matches]
        print(library_matches)
        library_samples = set([match[1] for match_list in library_matches for match in match_list])
        library_matched_functions = [key for key in self.library_matches if self.library_matches[key]]
        self.sample_matches = [sample_match for sample_match in self.sample_matches if sample_match.sample_id not in library_samples]
        self.function_matches = [function_match for function_match in self.function_matches if function_match.function_id not in library_matched_functions]
        self.is_library_filtered = True

    def filterToSampleCount(self, max_sample_count):
        """ reduce contained matches to those with a maximum of <max_sample_count> matched samples """
        matched_samples_by_function_id = {}
        for function_match in self.function_matches:
            if not function_match.function_id in matched_samples_by_function_id:
                matched_samples_by_function_id[function_match.function_id] = []
            if not function_match.matched_sample_id in matched_samples_by_function_id[function_match.function_id]:
                matched_samples_by_function_id[function_match.function_id].append(function_match.matched_family_id)
        self.function_matches = [function_match for function_match in self.function_matches if len(matched_samples_by_function_id[function_match.function_id]) <= max_sample_count]
        self.is_sample_count_filtered = True

    def filterToFamilyCount(self, max_family_count):
        """ reduce contained matches to those with a maximum of <max_family_count> matched families """
        matched_families_by_function_id = {}
        for function_match in self.function_matches:
            if not function_match.function_id in matched_families_by_function_id:
                matched_families_by_function_id[function_match.function_id] = []
            if not function_match.matched_family_id in matched_families_by_function_id[function_match.function_id]:
                matched_families_by_function_id[function_match.function_id].append(function_match.matched_family_id)
        self.function_matches = [function_match for function_match in self.function_matches if len(matched_families_by_function_id[function_match.function_id]) <= max_family_count]
        self.is_family_count_filtered = True

    def filterToScore(self, threshold):
        """ reduce contained matches to those with a minimum score of <threshold> """
        self.function_matches = [function_match for function_match in self.function_matches if function_match.matched_score >= threshold]
        self.is_score_filtered = True

    def filterToFamilyId(self, family_id):
        """ reduce contained matches to chosen family_id by deleting the other sample and function matches """
        self.sample_matches = [sample_match for sample_match in self.sample_matches if sample_match.family_id == family_id]
        self.function_matches = [function_match for function_match in self.function_matches if function_match.matched_family_id == family_id]
        self.is_family_filtered = True

    def filterToSampleId(self, sample_id):
        """ reduce contained matches to chosen sample_id by deleting the other sample and function matches """
        self.sample_matches = [sample_match for sample_match in self.sample_matches if sample_match.sample_id == sample_id]
        self.function_matches = [function_match for function_match in self.function_matches if function_match.matched_sample_id == sample_id]
        self.is_sample_filtered = True

    def filterToFunctionId(self, function_id):
        """ reduce contained matches to chosen function_id by deleting the other sample and function matches """
        # self.sample_matches = [sample_match for sample_match in self.sample_matches if sample_match.sample_id == sample_id]
        self.function_matches = [function_match for function_match in self.function_matches if function_match.function_id == function_id]
        self.is_function_filtered = True

    def getBestSampleMatchesPerFamily(self, start=None, limit=None):
        by_family = {}
        for sample_match in self.sample_matches:
            if sample_match.family not in by_family:
                by_family[sample_match.family] = {
                    "score": sample_match.matched_percent_frequency_weighted,
                    "report": sample_match
                }
            elif sample_match.matched_percent_frequency_weighted > by_family[sample_match.family]["score"]:
                by_family[sample_match.family]["score"] = sample_match.matched_percent_frequency_weighted
                by_family[sample_match.family]["report"] = sample_match
        result_list = []
        for family, score_entry in sorted(by_family.items(), key=lambda e: e[1]["score"], reverse=True):
            result_list.append(score_entry["report"])
        if start is not None:
            result_list = result_list[start:]
        if limit is not None:
            result_list = result_list[:limit]
        return result_list

    def getUniqueFamilyMatchInfoForSample(self, sample_id):
        if self.unique_family_scores_per_sample is None:
            self.unique_family_scores_per_sample = {entry.sample_id: {"functions_matched": 0, "bytes_matched": 0, "unique_score": 0} for entry in self.sample_matches}
            families_matched_by_function_id = {}
            samples_matched_by_function_id = {}
            bytes_per_function_id = {}
            for function_match_summary in self.function_matches:
                if function_match_summary.function_id not in families_matched_by_function_id:
                    families_matched_by_function_id[function_match_summary.function_id] = set()
                    samples_matched_by_function_id[function_match_summary.function_id] = set()
                    bytes_per_function_id[function_match_summary.function_id] = 0
                families_matched_by_function_id[function_match_summary.function_id].add(function_match_summary.matched_family_id)
                samples_matched_by_function_id[function_match_summary.function_id].add(function_match_summary.matched_sample_id)
                bytes_per_function_id[function_match_summary.function_id] = function_match_summary.num_bytes
            for function_id in families_matched_by_function_id:
                if len(families_matched_by_function_id[function_id]) == 1:
                    for sample_id in samples_matched_by_function_id[function_id]:
                        self.unique_family_scores_per_sample[sample_id]["functions_matched"] += 1
                        self.unique_family_scores_per_sample[sample_id]["bytes_matched"] += bytes_per_function_id[function_id]
            for sample_id in self.unique_family_scores_per_sample:
                self.unique_family_scores_per_sample[sample_id]["unique_score"] = 100.0 * self.unique_family_scores_per_sample[sample_id]["bytes_matched"] / self.reference_sample_entry.binweight
        return self.unique_family_scores_per_sample[sample_id]

    def getSampleMatches(self, start=None, limit=None):
        by_sample_id = {}
        for sample_match in self.sample_matches:
            by_sample_id[sample_match.sample_id] = {
                "score": sample_match.matched_percent_frequency_weighted,
                "report": sample_match
            }
        result_list = []
        for sample_id, score_entry in sorted(by_sample_id.items(), key=lambda e: e[1]["score"], reverse=True):
            result_list.append(score_entry["report"])
        if start is not None:
            result_list = result_list[start:]
        if limit is not None:
            result_list = result_list[:limit]
        return result_list

    def getAggregatedFunctionMatches(self, start=None, limit=None):
        by_function_id = {}
        for function_match in self.function_matches:
            if function_match.function_id not in by_function_id:
                by_function_id[function_match.function_id] = {
                    "function_id": function_match.function_id,
                    "num_bytes": 0,
                    "offset": 0,
                    "best_score": 0,
                    "num_families_matched": 0,
                    "family_ids_matched": set([]),
                    "num_samples_matched": 0,
                    "sample_ids_matched": set([]),
                    "num_functions_matched": 0,
                    "function_ids_matched": set([]),
                    "minhash_matches": 0,
                    "pichash_matches": 0,
                    "library_matches": 0,
                }
            by_function_id[function_match.function_id]["num_bytes"] = function_match.num_bytes
            by_function_id[function_match.function_id]["offset"] = function_match.offset
            by_function_id[function_match.function_id]["best_score"] = max(function_match.matched_score, by_function_id[function_match.function_id]["best_score"])
            by_function_id[function_match.function_id]["family_ids_matched"].add(function_match.matched_family_id)
            by_function_id[function_match.function_id]["sample_ids_matched"].add(function_match.matched_sample_id)
            by_function_id[function_match.function_id]["function_ids_matched"].add(function_match.matched_function_id)
            by_function_id[function_match.function_id]["num_families_matched"] = len(by_function_id[function_match.function_id]["family_ids_matched"])
            by_function_id[function_match.function_id]["num_samples_matched"] = len(by_function_id[function_match.function_id]["sample_ids_matched"])
            by_function_id[function_match.function_id]["num_functions_matched"] = len(by_function_id[function_match.function_id]["function_ids_matched"])
            by_function_id[function_match.function_id]["minhash_matches"] += 1 if function_match.match_is_minhash else 0
            by_function_id[function_match.function_id]["pichash_matches"] += 1 if function_match.match_is_pichash else 0
            by_function_id[function_match.function_id]["library_matches"] += 1 if function_match.match_is_library else 0
        aggregated_matched = [v for k, v in sorted(by_function_id.items())]
        if start is not None:
            aggregated_matched = aggregated_matched[start:]
        if limit is not None:
            aggregated_matched = aggregated_matched[:limit]
        return aggregated_matched

    def getFunctionsSlice(self, start, limit):
        return self.function_matches[start:start+limit]

    def toDict(self):
        # we need to aggregate by function_id here
        summarized_function_match_summaries = {}
        for function_match_entry in self.function_matches:
            if function_match_entry.function_id not in summarized_function_match_summaries:
                summarized_function_match_summaries[function_match_entry.function_id] = {
                    "num_bytes": function_match_entry.num_bytes,
                    "offset": function_match_entry.offset,
                    "fid": function_match_entry.function_id,
                    "matches": [function_match_entry.getMatchTuple()]
                }
            else:
                summarized_function_match_summaries[function_match_entry.function_id]["matches"].append(function_match_entry.getMatchTuple())
        # build the dictionary
        matching_entry = {
            "info": {
                "sample": self.reference_sample_entry.toDict()
            },
            "matches": {
                "aggregation": self.match_aggregation,
                "functions": summarized_function_match_summaries,
                "samples": [match.toDict() for match in self.sample_matches]
            }
        }
        if self.other_sample_entry is not None:
            matching_entry["other_sample_info"] = self.other_sample_entry.toDict()
        return matching_entry

    @classmethod
    def fromDict(cls, entry_dict):
        matching_entry = cls(None)
        matching_entry.reference_sample_entry = SampleEntry.fromDict(entry_dict["info"]["sample"])
        if "other_sample_info" in entry_dict:
            matching_entry.other_sample_entry = SampleEntry.fromDict(entry_dict["other_sample_info"])
        else:
            matching_entry.other_sample_entry = None
        matching_entry.match_aggregation = entry_dict["matches"]["aggregation"]
        matching_entry.sample_matches = [MatchedSampleEntry.fromDict(entry) for entry in entry_dict["matches"]["samples"]]
        # expand function matches into individual entries
        list_of_function_matches = []
        matching_entry.library_matches = {entry["fid"]: [] for entry in entry_dict["matches"]["functions"]}
        matching_entry.unique_family_scores_per_sample = None
        for function_match_summary in entry_dict["matches"]["functions"]:
            num_bytes = function_match_summary["num_bytes"]
            offset = function_match_summary["offset"]
            function_id = function_match_summary["fid"]
            for match_tuple in function_match_summary["matches"]:
                list_of_function_matches.append(MatchedFunctionEntry(function_id, num_bytes, offset, match_tuple))
                if match_tuple[4] & MatcherInterface.IS_LIBRARY_FLAG:
                    if (match_tuple[0], match_tuple[1]) not in matching_entry.library_matches[function_id]:
                        matching_entry.library_matches[function_id].append((match_tuple[0], match_tuple[1]))
        matching_entry.function_matches = list_of_function_matches
        return matching_entry

    def __str__(self):
        return "Matched: Samples: {} Functions: {}".format(
            len(self.sample_matches),
            len(self.function_matches),
        )



from typing import TYPE_CHECKING, Dict, List, Optional

from mcrit.storage.SampleEntry import SampleEntry

if TYPE_CHECKING:  # pragma: no cover
    from mcrit.storage.SampleEntry import SampleEntry

# Dataclass, post init
# constructor -> .fromSmdaFunction
# assume sample_entry, smda_function always available

class MatchedSampleEntry(object):
    # basic information
    family: str
    family_id: int
    version: str
    bitness: int
    sha256: str
    filename: str
    sample_id: int
    num_bytes: int
    num_functions: int
    # related to matching
    matched_functions_minhash: int
    matched_functions_pichash: int
    matched_functions_combined: int
    matched_functions_library: int
    # scores
    matched_bytes_unweighted: float
    matched_bytes_score_weighted: float
    matched_bytes_frequency_weighted: float
    matched_bytes_nonlib_unweighted: float
    matched_bytes_nonlib_score_weighted: float
    matched_bytes_nonlib_frequency_weighted: float
    matched_percent_unweighted: float
    matched_percent_score_weighted: float
    matched_percent_frequency_weighted: float
    matched_percent_nonlib_unweighted: float
    matched_percent_nonlib_score_weighted: float
    matched_percent_nonlib_frequency_weighted: float

    def __init__(self, sample_id: int) -> None:
        self.sample_id = sample_id


    def getShortSha256(self, prefix=8, border=0):
        if border > 0:
            return self.sha256[:border] + "..." + self.sha256[-border:]
        elif prefix > 0:
            return self.sha256[:prefix]
        return self.sha256

    def getShortFilename(self, size_visible=20):
        if (len(self.filename) > 2 * size_visible):
            return self.filename[:size_visible] + "..." + self.filename[-size_visible:]
        return self.filename

    def toDict(self):
        matching_entry = {
            "family": self.family,
            "family_id": self.family_id,
            "version": self.version,
            "bitness": self.bitness,
            "sha256": self.sha256,
            "filename": self.filename,
            "sample_id": self.sample_id,
            "num_bytes": self.num_bytes,
            "num_functions": self.num_functions,
            "matched": {
                "functions": {
                    "minhashes": self.matched_functions_minhash,
                    "pichashes": self.matched_functions_pichash,
                    "combined": self.matched_functions_combined,
                    "library": self.matched_functions_library,
                },
                "bytes": {
                    "unweighted": self.matched_bytes_unweighted,
                    "score_weighted": self.matched_bytes_score_weighted,
                    "frequency_weighted": self.matched_bytes_frequency_weighted,
                    "nonlib_unweighted": self.matched_bytes_nonlib_unweighted,
                    "nonlib_score_weighted": self.matched_bytes_nonlib_score_weighted,
                    "nonlib_frequency_weighted": self.matched_bytes_nonlib_frequency_weighted
                },
                "percent": {
                    "unweighted": self.matched_percent_unweighted,
                    "score_weighted": self.matched_percent_score_weighted,
                    "frequency_weighted": self.matched_percent_frequency_weighted,
                    "nonlib_unweighted": self.matched_percent_nonlib_unweighted,
                    "nonlib_score_weighted": self.matched_percent_nonlib_score_weighted,
                    "nonlib_frequency_weighted": self.matched_percent_nonlib_frequency_weighted
                }
            }
        }
        return matching_entry

    @classmethod
    def fromDict(cls, entry_dict):
        matching_entry = cls(None)
        matching_entry.family = entry_dict["family"]
        matching_entry.family_id = entry_dict["family_id"]
        matching_entry.version = entry_dict["version"]
        matching_entry.bitness = entry_dict["bitness"]
        matching_entry.sha256 = entry_dict["sha256"]
        matching_entry.filename = entry_dict["filename"]
        matching_entry.sample_id = entry_dict["sample_id"]
        matching_entry.num_bytes = entry_dict["num_bytes"]
        matching_entry.num_functions = entry_dict["num_functions"]

        matching_entry.matched_functions_minhash = entry_dict["matched"]["functions"]["minhashes"]
        matching_entry.matched_functions_pichash = entry_dict["matched"]["functions"]["pichashes"]
        matching_entry.matched_functions_combined = entry_dict["matched"]["functions"]["combined"]
        matching_entry.matched_functions_library = entry_dict["matched"]["functions"]["library"]

        matching_entry.matched_bytes_unweighted = entry_dict["matched"]["bytes"]["unweighted"]
        matching_entry.matched_bytes_score_weighted = entry_dict["matched"]["bytes"]["score_weighted"]
        matching_entry.matched_bytes_frequency_weighted = entry_dict["matched"]["bytes"]["frequency_weighted"]
        matching_entry.matched_bytes_nonlib_unweighted = entry_dict["matched"]["bytes"]["nonlib_unweighted"]
        matching_entry.matched_bytes_nonlib_score_weighted = entry_dict["matched"]["bytes"]["nonlib_score_weighted"]
        matching_entry.matched_bytes_nonlib_frequency_weighted = entry_dict["matched"]["bytes"]["nonlib_frequency_weighted"]

        matching_entry.matched_percent_unweighted = entry_dict["matched"]["percent"]["unweighted"]
        matching_entry.matched_percent_score_weighted = entry_dict["matched"]["percent"]["score_weighted"]
        matching_entry.matched_percent_frequency_weighted = entry_dict["matched"]["percent"]["frequency_weighted"]
        matching_entry.matched_percent_nonlib_unweighted = entry_dict["matched"]["percent"]["nonlib_unweighted"]
        matching_entry.matched_percent_nonlib_score_weighted = entry_dict["matched"]["percent"]["nonlib_score_weighted"]
        matching_entry.matched_percent_nonlib_frequency_weighted = entry_dict["matched"]["percent"]["nonlib_frequency_weighted"]

        return matching_entry

    def __str__(self):
        return "Matched Sample: id({}) family({}) - Matched: {} ({}) - {} ({}) {} ({})".format(
            self.sample_id,
            self.family,
            self.matched_bytes_unweighted,
            self.matched_percent_unweighted,
            self.matched_bytes_score_weighted,
            self.matched_percent_score_weighted,
            self.matched_bytes_frequency_weighted,
            self.matched_percent_frequency_weighted,
        )
